#ifndef CALCULATOR_FUNCTIONS_H_INCLUDED
#define CALCULATOR_FUNCTIONS_H_INCLUDED

#include <iostream>
#include <optional>

#include "configuration.h"
#include "payment_method.h"
#include "ticket.h"
#include "tariff_time_range.h"

#include <QDateTime>
using namespace std;

class Calculator
{
public:

	/// <summary>
	/// Gets duration in seconds from cost
	/// </summary>
	/// <param name="tariff_cfg">Pointer to configuration</param>
	/// <param name="vehicle_type">Type of vehicle</param>
	/// <param name="start_datetime">Date/time of payment to be conducted in ISO8601 format (e.g. 2022-12-25T08:00:00Z)</param>
	/// <param name="price"></param>
	/// <returns>Returns duration in seconds (data type: double)</returns>
	std::string GetDurationFromCost(Configuration* cfg, uint8_t vehicle_type, char const* start_datetime, double price, bool nextDay = false, bool prepaid = false);

	/// <summary>
	///  Gets cost from duration in seconds
	/// </summary>
	/// <param name="tariff_cfg">Pointer to configuration</param>
	/// <param name="vehicle_type">Type of vehicle</param>
	/// <param name="start_datetime">Date/time of payment to be conducted in ISO8601 format (e.g. 2022-12-25T08:00:00Z) </param>
    /// <param name="end_datetime">Date/time of park end to be conducted in ISO8601 format (e.g. 2022-12-25T08:00:00Z) </param>
	/// <param name="durationMin">Duration of parking in minutes</param>
	/// <returns>Returns cost (data type: double)</returns>
    double GetCostFromDuration(Configuration* cfg, uint8_t vehicle_type, const QDateTime start_datetime, QDateTime & end_datetime, int durationMin, bool nextDay = false, bool prepaid = false);

    // Daily ticket
    QDateTime GetDailyTicketDuration(Configuration* cfg, const QDateTime start_datetime, uint8_t payment_option, bool carry_over);

    //
    // helper function to find time steps for a tariff with PaymentMethod::Steps
    // (e.g. Schoenau/Koenigsee)
    //
    QList<int> GetTimeSteps(Configuration *cfg) const;

private:
    // Introduced for PaymentMethod::Steps (e.g. Schoenau)
    // For tariff of following structure: only steps, no special days, nonstop.
    uint32_t GetCostFromDuration(Configuration *cfg, QDateTime const &start, quint64 durationMinutes) const;
    uint32_t GetCostFromDuration(Configuration *cfg, QDateTime const &start, QDateTime const &end) const;

    PaymentMethod getPaymentMethodId(Configuration const *cfg);
    int getMinimalParkingTime(Configuration const *cfg, PaymentMethod methodId);
    int getMaximalParkingTime(Configuration const *cfg, PaymentMethod methodId);
    uint32_t getMinimalParkingPrice(Configuration const *cfg, PaymentMethod methodId);

    Ticket private_GetCostFromDuration(Configuration const* cfg,
                                       QDateTime const &start,
                                       int durationMinutes,
                                       bool prepaid = false);
    Ticket private_GetDurationFromCost(Configuration *cfg,
                                       QDateTime const &start,
                                       uint32_t price,
                                       bool prepaid = false);

    bool checkDurationMinutes(int minParkingTime, int maxParkingTime,
                              int durationMinutes);

    //
    uint32_t GetPriceForTimeStep(Configuration *cfg, int timeStep) const;
    uint32_t GetDurationForPrice(Configuration *cfg, int price) const;

    int findWorkTimeRange(QDateTime const &dt,
                          QScopedArrayPointer<TariffTimeRange> const &worktime,
                          size_t size);
    int findNextWorkTimeRange(QDateTime const &dt,
                              QScopedArrayPointer<TariffTimeRange> const &worktime,
                              size_t size);

    uint32_t computeWeekDaysPrice(Configuration const *cfg, PaymentMethod id) const;
    double computeWeekDaysDurationUnit(Configuration const *cfg, PaymentMethod id) const;
};

#endif // CALCULATOR_FUNCTIONS_H_INCLUDED
